#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Créé Déc 2024

@ auteur : David ALBERTO (www.astrolabe-science.fr)
Trace la limite jour-nuit sur une carte de France, pour différentes dates.
L'heure est déterminée comme l'heure de coucher du Soleil en un point
choisi par ses coordonnées géographiques, ou bien le centre de la France.
L'heure du coucher calculée est valable à 5 min près.
"""
import matplotlib.pyplot as plt
import numpy as np
import geopandas as gpd
import datetime as dt
from shapely.ops import unary_union
from math import sin, asin, degrees, radians
import locale  # réglages locaux de langue

#  -------------------------- PERSONNALISATION -----------------------

plt.rcParams["font.family"] = "Roboto"
plt.rcParams["font.size"] = 10
opacite = 0.6 # entre 0 et 1, pour la zone "nuit".
couleur_centre = 'crimson'
locale.setlocale(locale.LC_ALL, 'fr_FR.UTF-8')  # dates en français

liste_dates = [
    #  rappel : le format est : (année, mois, jour, heure, minute, seconde)
    dt.datetime(2024, 1, 20, 17),
    dt.datetime(2024, 2, 20, 17),
    dt.datetime(2024, 3, 20, 18),
    dt.datetime(2024, 4, 20, 18),
    dt.datetime(2024, 5, 20, 18),
    dt.datetime(2024, 6, 21, 20),
    dt.datetime(2024, 7, 20, 19, 30),
    dt.datetime(2024, 8, 20, 19),
    dt.datetime(2024, 9, 22, 19),
    dt.datetime(2024, 10, 20, 19),
    dt.datetime(2024, 11, 20, 16,30),
    dt.datetime(2024, 12, 21, 16),
    ]

eps = np.radians(23.439)  # obliquité de l'écliptique (calcul de déclinaison)

#  -------------------------- FONCTIONS -----------------------

def jour_mois_an(date):
    """
    date : date au format datetime.
    Renvoie les numéros du jour, du mois et de l'année (entiers).
    """
    j = date.day
    m = date.month
    a = date.year
    return j, m, a

def declinaison(date, deg = False):
    """
    D'après D. Savoie (Gnomonique).
    Parameters
    ----------
    date : date au format datetime
    deg : booléen (True si on veut la valeur en degrés)
    Returns
    -------
    Renvoie la déclinaison du Soleil à la date demandée (radians par défaut, 
                                                         en degrés si deg=True).
    """
    Tzero = dt.datetime(2000, 1, 1, 12) #  1er janvier 2000 12hTU, date de référence
    T = (date - Tzero).days
    T = T + (date - Tzero).seconds / 3600 / 24
    LM = (280.460 + 0.9856474 * T) % 360
    M = (357.528 + 0.9856003 * T) % 360
    long = (LM + 1.915 * sin(radians(M)) + 0.020 * sin(radians(2*M))) % 360
    declinaison = asin(sin(eps) * sin(radians(long)))
    if deg:
        declinaison = degrees(declinaison)
    return declinaison

def search_centroid():
    """
    Recherche le centroïde de la France (dont Corse).
    Renvoie longitude et latitude.
    """
    #  recherche du centroide de la France :
    polygones = france.geometry
    union = gpd.GeoSeries(unary_union(polygones))
    x, y = union.centroid.x[0], union.centroid.y[0]
    return x, y

def heure_coucher_point(t):
    """
    t : date format datetime.
    Calcule l'heure de coucher du Soleil au point de référence, pour la date t.
    Affiche l'heure légale de coucher.
    Renvoie l'heure solaire (float) et l'arc semi-diurne (float en degrés).
    """
    cosH0 = -np.tan(np.radians(y_point)) * np.tan(decl_rad)
    H0 = np.arccos(cosH0)
    edt = equation_temps(date)[0]
    coucher_sol = 12 + H0 / np.radians(15)  # heure solaire du coucher
    # fuseau horaire français (dates de changement d'heure approx. !):
    if t >= dt.datetime(2024, 3, 30) and t <= dt.datetime(2024, 9, 30):
        F = 2.0
    else:
        F = 1.0
    coucher_leg = coucher_sol + edt/60 - x_point * 4 / 60 + F
    # affichage de l'heure légale de coucher au point de référence :
    ax.text(0.01, 0.01, HHmm(coucher_leg), c=couleur_centre,
            transform = ax.transAxes)
    return coucher_sol, np.degrees(H0)

def HHmm(H):
    """
    Transforme l'heure décimale en format sexagésimal.
    Parameters
    ----------
    H : TYPE : float
        DESCRIPTION : heure en format décimal.
    -------
    Returns : heure en format string (heure:min)
    """
    heures = int(H)
    minutes = round((H-heures)*60)
    if minutes == 60:
        minutes = '00'
        heures = heures + 1
    elif minutes <10:
        minutes = '0' + str(minutes)
    heure_60 = f"{heures}h{minutes}min"
    return heure_60

def equation_temps(date):
    """
    Calcule l'équation du temps pour une date donnée.
    D'après D. Savoie (Gnomonique) Précision env.30 s pour 1950-2050.
    Parameters
    ----------
    date : TYPE : date au format datetime : dt.datetime(annee, mois , jour)
        DESCRIPTION : date du jour souhaité.
    Returns
    -------
    Renvoie la valeur de l'équation du temps à la date demandée,
    sous forme de tuple : (valeur décimale, string en format min:sec).
    """
    Tzero = dt.datetime(2000, 1, 1, 12) #  1er janvier 2000 12hTU, date de référence
    # date = dt.datetime(annee, mois, jour)
    T = (date - Tzero).days
    T = T + (date - Tzero).seconds / 3600 / 24
    LM = (280.460 + 0.9856474 * T) % 360
    M = (357.528 + 0.9856003 * T) % 360
    long = (LM + 1.915 * np.sin(np.radians(M)) + 0.020 * np.sin(np.radians(2*M))) % 360
    AD = np.arctan2(np.cos(eps) * np.sin(np.radians(long)), np.cos(np.radians(long)))
    EdT_degres = ((np.degrees(AD) - LM))%360
    if EdT_degres > 180:
        EdT_degres = EdT_degres - 360
    EdT = EdT_degres * 4
    minutes = int(EdT)
    secondes = round((abs(EdT - minutes)) * 60,1)
    return EdT, f"{minutes}min{secondes}s"

def trace_ombre(t):
    """
    Trace la limite d'ombre pour la date donnée.
    """
    H0 = heure_coucher_point(date)[1]
    # calculs des latitudes de la courbe d'ombre :

    # decl_rad = np.radians(decl_deg)
    longitudes = np.arange(xmin, xmax+0.01, 0.01)
    H_deg = H0 - x_point + longitudes  # angles horaires des longitudes
    H_rad = np.radians(H_deg)
    # calcul des latitudes où le Soleil est à l'horizon :
    latitudes_deg = np.degrees(np.arctan(-np.cos(H_rad)/np.tan(decl_rad)))
    #  courbe de limite jour/nuit :
    x = longitudes
    y = latitudes_deg
    ax.plot(x, y, c='k', lw=0.5)
    # affichage de la date :
    ax.text(0.01, 0.99, t.strftime('%d %b'), c= 'indigo', va='top',
            transform = ax.transAxes, fontsize=12)

#  zone d'ombre :
    if decl_deg == 0.0 :  # si équinoxes
        x = np.arange(x_point, xmax+0.01, 0.01)
        ax.fill_between(x, ymin, ymax, color='black', alpha=opacite)
    elif decl_deg > 0.0:
        ax.fill_between(x, y, ymin, color='black', alpha=opacite)
    else :
        ax.fill_between(x, y, ymax, color='black', alpha=opacite)

# ------------------------------------------------------------------
#  paramètres du graphique :

fig = plt.figure(tight_layout=True, figsize=(5,5))

ax = plt.subplot(111)
# ------------------------------------------------------------------
    # import des données SIG des départements :
france = gpd.read_file('departements.geojson')

    # définition du point de référence :

x_point, y_point = search_centroid()  # centroide de la carte
# x_point, y_point = 0.11, 49.5  # longitude et latitude entrées manuellement
# x_point, y_point = 2.3, 48.85  # longitude et latitude entrées manuellement


#  tracés :

for N, date in enumerate(liste_dates):
    # jour, mois, an = jour_mois_an(date)
    ax.clear()
    france.plot(ax=ax, facecolor='lightsteelblue', edgecolor='indigo', lw=0.1)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    ax.scatter(x_point, y_point, s=10, marker='o', color=couleur_centre, lw=0.4)

#  ---------------------------------------------------------

    decl_rad = declinaison(date, deg=False)
    decl_deg = np.degrees(decl_rad)
    # decl_rad = np.radians(decl_deg)

    trace_ombre(date)
    
        # limites des axes :
    ax.set(
            xlim = (xmin, xmax),
            ylim = (ymin, ymax),
            xticks=[],
            yticks=[],
            )

    fig.savefig(f'limite_jour_nuit_France{10+N}.png', dpi=200)
